"""add downgrade_after

Revision ID: 18b0a0874c5a
Revises: d363aeb3c4de
Create Date: 2016-09-28 03:00:45.197012

"""

# revision identifiers, used by Alembic.
revision = '18b0a0874c5a'
down_revision = 'd363aeb3c4de'

from alembic import op
import sqlalchemy as sa


def upgrade():
    ### commands auto generated by Alembic - please adjust! ###
    op.add_column('release_monitor', sa.Column('downgrade_after', sa.Integer(), nullable=True))
    op.add_column('release_monitor', sa.Column('last_torrent_update_stamp', sa.Integer(), nullable=True))
    op.add_column('show', sa.Column('downgrade_after', sa.Integer(), nullable=True))
    op.add_column('torrent', sa.Column('caching', sa.Boolean(), nullable=True))
    ### end Alembic commands ###


def downgrade():
    ### commands auto generated by Alembic - please adjust! ###
    op.add_column('release_monitor', sa.Column('torrent_cached', sa.BOOLEAN(), nullable=True))
    op.add_column('link', sa.Column('real_type', sa.VARCHAR(length=24), nullable=False))
    ### end Alembic commands ###
